import numpy as np
import soundfile as sf
import pandas as pd
import os
import scipy.stats

def compute_spectral_clarity(segment_data, threshold=0.85):
    # Compute the spectral clarity of a segment of audio data.
    # Spectral clarity is defined as the mean value of the top portion of the spectrum,
    # determined by a threshold.
    spectrum = np.abs(np.fft.fft(segment_data))
    sorted_spectrum = np.sort(spectrum)[::-1]
    cumulative_energy = np.cumsum(sorted_spectrum)
    total_energy = cumulative_energy[-1]
    clarity_index = np.where(cumulative_energy >= threshold * total_energy)[0][0]
    spectral_clarity = sorted_spectrum[:clarity_index].mean()
    return spectral_clarity

def compute_spectral_energy_distribution(segment_data, window_size=1024):
    # Compute the spectral energy distribution over windows of the given size.
    # This is a measure of how energy is distributed across different frequencies.
    windowed_data = np.lib.stride_tricks.sliding_window_view(segment_data, window_size)
    energy_distribution = np.sum(windowed_data**2, axis=1)
    return energy_distribution

def compute_spectral_tilt(segment_data, samplerate):
    # Calculate the spectral tilt, which is a measure of the slope of the spectrum.
    # This can give an idea of the balance between high and low frequencies.
    spectrum = np.abs(np.fft.fft(segment_data))
    frequencies = np.fft.fftfreq(len(segment_data), 1/samplerate)
    spectral_tilt = np.sum(frequencies * spectrum) / np.sum(spectrum)
    return spectral_tilt

def compute_spectral_flatness(segment_data):
    # Compute the spectral flatness, a measure that indicates how noise-like a signal is,
    # as opposed to being tonal. A higher flatness indicates a more noise-like signal.
    spectrum = np.abs(np.fft.fft(segment_data))
    geometric_mean = scipy.stats.gmean(spectrum)
    arithmetic_mean = np.mean(spectrum)
    spectral_flatness = geometric_mean / arithmetic_mean
    return spectral_flatness

def process_audio_channels(data):
    # Process the audio channels (stereo to mono conversion) by averaging the channels,
    # weighted by their average absolute amplitude.
    if len(data.shape) == 2:
        left_channel = data[:, 0]
        right_channel = data[:, 1]
        weight_left = np.mean(np.abs(left_channel))
        weight_right = np.mean(np.abs(right_channel))
        return (left_channel * weight_right + right_channel * weight_left) / (weight_left + weight_right)
    return data

def compute_rolling_amplitude_avg(data, samplerate):
    # Compute the rolling average of amplitude over a window (80 ms in this case),
    # which can be useful for identifying different segments of the audio.
    window_size = int(0.08 * samplerate)  # 80 ms window
    return np.convolve(np.abs(data), np.ones(window_size)/window_size, mode='valid')

def find_audio_segments(amplitude_rolling_avg, samplerate):
    # Identify different segments of the audio based on amplitude changes.
    # This includes determining the start, stable, and end segments.
    start_threshold = 0.70 * np.max(amplitude_rolling_avg)
    start_idx = next((i for i, amp in enumerate(amplitude_rolling_avg) if amp > start_threshold), 0)

    end_threshold_low = 0.1 * np.max(amplitude_rolling_avg)
    end_threshold_high = 0.99 * np.max(amplitude_rolling_avg)

    consecutive_count = 0
    for i, amp in enumerate(amplitude_rolling_avg[start_idx:]):
        if amp < end_threshold_low:
            consecutive_count += 1
            if consecutive_count >= 5:
                end_temp_idx = i - 4
                break
        else:
            consecutive_count = 0
    else:
        end_temp_idx = len(amplitude_rolling_avg) - start_idx

    end_idx = start_idx + end_temp_idx - next((i for i in range(end_temp_idx, 0, -1) if start_idx + i < len(amplitude_rolling_avg) and amplitude_rolling_avg[start_idx + i] > end_threshold_high), end_temp_idx)

    segments = {
        "Start": (0, start_idx),
        "Stable": (start_idx, end_idx),
        "End": (end_idx, len(data))
    }
    return segments

def calculate_spectral_centroid(data, samplerate, segments):
    # Calculate the spectral centroid for the stable segment of the audio.
    # The spectral centroid is a measure that indicates where the center of mass of the spectrum is located.
    for name, (start, end) in segments.items():
        if name == "Stable":
            segment_data = data[start:end]
            spectrum = np.abs(np.fft.fft(segment_data))
            frequencies = np.fft.fftfreq(len(segment_data), 1/samplerate)
            positive_freq_indices = np.where(frequencies > 0)
            spectrum = spectrum[positive_freq_indices]
            frequencies = frequencies[positive_freq_indices]
            spectral_centroid = np.sum(frequencies * spectrum) / np.sum(spectrum)

            time_start = start / samplerate
            time_end = end / samplerate

            return spectral_centroid, (time_start, time_end)


def read_audio_files(folder_path):
    # Read audio files from a given folder path.
    return [file for file in os.listdir(folder_path) if file.endswith('.wav')]

def save_results_to_excel(results):
    # Save the computed results to an Excel file.
    df = pd.DataFrame(results)
    df.to_excel('output_spectral_centroids.xlsx', index=False)

# Main function or script part
folder_path = '/Users/xujiasheng/Desktop/test/records'
audio_files = read_audio_files(folder_path)

results = []

for audio_file in audio_files:
    data, samplerate = sf.read(os.path.join(folder_path, audio_file))
    data = process_audio_channels(data)
    amplitude_rolling_avg = compute_rolling_amplitude_avg(data, samplerate)
    segments = find_audio_segments(amplitude_rolling_avg, samplerate)
    spectral_centroid, time_segment = calculate_spectral_centroid(data, samplerate, segments)

    # Compute additional features for the 'stable' segment
    if 'Stable' in segments:
        start, end = segments['Stable']
        stable_segment = data[start:end]

        spectral_flatness = compute_spectral_flatness(stable_segment)
        spectral_tilt = compute_spectral_tilt(stable_segment, samplerate)
        spectral_clarity = compute_spectral_clarity(stable_segment)
        energy_distribution = compute_spectral_energy_distribution(stable_segment)

        results.append({
            'Audio_File': audio_file,
            'Spectral_Centroid': spectral_centroid,
            'Spectral_Flatness': spectral_flatness,
            'Spectral_Tilt': spectral_tilt,
            'Spectral_Clarity': spectral_clarity,
            'Energy_Distribution': energy_distribution.tolist(),  # Convert array to list
            'Time_Segment': time_segment
        })

save_results_to_excel(results)
